@@ -8,7 +8,7 @@ typedef enum sr_error_e {
     SR_ERR_INVAL_ARG,
     SR_ERR_LY,
     SR_ERR_SYS,
-    SR_ERR_NOMEM,
+    SR_ERR_NO_MEMORY,
     SR_ERR_NOT_FOUND,
     SR_ERR_EXISTS,
     SR_ERR_INTERNAL,
@@ -45,8 +45,7 @@ typedef struct sr_conn_ctx_s sr_conn_ctx_t;
 typedef struct sr_session_ctx_s sr_session_ctx_t;
 typedef enum sr_conn_flag_e {
 	SR_CONN_CACHE_RUNNING,
-	SR_CONN_NO_SCHED_CHANGES,
-	SR_CONN_ERR_ON_SCHED_FAIL,
+	SR_CONN_CTX_SET_PRIV_PARSED,
 	...
 } sr_conn_flag_t;
 typedef uint32_t sr_conn_options_t;
@@ -57,25 +56,32 @@ typedef enum sr_datastore_e {
 	SR_DS_OPERATIONAL,
 	...
 } sr_datastore_t;
-typedef struct sr_error_info_msg_s {
-	char *message;
-	char *xpath;
-} sr_error_info_msg_t;
-typedef struct sr_error_info_s {
+typedef struct sr_error_info_err_s {
 	sr_error_t err_code;
-	sr_error_info_msg_t *err;
-	size_t err_count;
+	char *message;
+	char *error_format;
+	void *error_data;
+} sr_error_info_err_t;
+typedef struct sr_error_info_s {
+	sr_error_info_err_t *err;
+	uint32_t err_count;
 } sr_error_info_t;
 
 /* forward declarations from libyang */
 struct ly_ctx;
 struct lyd_node;
+typedef int... time_t;
+struct timespec {
+    time_t tv_sec;
+    long tv_nsec;
+};
 
 int sr_connect(const sr_conn_options_t, sr_conn_ctx_t **);
 int sr_disconnect(sr_conn_ctx_t *);
-const struct ly_ctx *sr_get_context(sr_conn_ctx_t *);
-int sr_install_module(sr_conn_ctx_t *, const char *, const char *, const char **, int);
-int sr_remove_module(sr_conn_ctx_t *, const char *);
+const struct ly_ctx *sr_acquire_context(sr_conn_ctx_t *);
+void sr_release_context(sr_conn_ctx_t *);
+int sr_install_module(sr_conn_ctx_t *, const char *, const char *, const char **);
+int sr_remove_module(sr_conn_ctx_t *, const char *, int);
 int sr_update_module(sr_conn_ctx_t *, const char *, const char *);
 int sr_enable_module_feature(sr_conn_ctx_t *, const char *, const char *);
 int sr_disable_module_feature(sr_conn_ctx_t *, const char *, const char *);
@@ -85,12 +91,14 @@ int sr_session_stop(sr_session_ctx_t *);
 int sr_session_switch_ds(sr_session_ctx_t *, sr_datastore_t);
 sr_datastore_t sr_session_get_ds(sr_session_ctx_t *);
 sr_conn_ctx_t *sr_session_get_connection(sr_session_ctx_t *);
-uint32_t sr_session_get_event_nc_id(sr_session_ctx_t *);
-const char *sr_session_get_event_user(sr_session_ctx_t *);
-int sr_get_error(sr_session_ctx_t *, const sr_error_info_t **);
-int sr_set_error(sr_session_ctx_t *, const char *, const char *, ...);
+int sr_session_get_error(sr_session_ctx_t *, const sr_error_info_t **);
+int sr_session_set_error_message(sr_session_ctx_t *, const char *, ...);
+const char *sr_session_get_orig_name(sr_session_ctx_t *session);
+int sr_session_set_orig_name(sr_session_ctx_t *session, const char *);
+int sr_session_get_orig_data(sr_session_ctx_t *session, uint32_t idx, uint32_t *size, const void **data);
+int sr_session_push_orig_data(sr_session_ctx_t * session, uint32_t size, const void *data);
 
-typedef enum sr_type_e {
+typedef enum sr_val_type_e {
 	SR_UNKNOWN_T,
 	SR_LIST_T,
 	SR_CONTAINER_T,
@@ -115,11 +123,11 @@ typedef enum sr_type_e {
 	SR_ANYXML_T,
 	SR_ANYDATA_T,
 	...
-} sr_type_t;
-typedef union sr_data_u {
+} sr_val_type_t;
+typedef union sr_val_data_u {
 	char *binary_val;
 	char *bits_val;
-	bool bool_val;
+	int bool_val;
 	double decimal64_val;
 	char *enum_val;
 	char *identityref_val;
@@ -136,11 +144,11 @@ typedef union sr_data_u {
 	char *anyxml_val;
 	char *anydata_val;
 	...;
-} sr_data_t;
+} sr_val_data_t;
 typedef struct sr_val_s {
     char *xpath;
-    sr_type_t type;
-    sr_data_t data;
+    sr_val_type_t type;
+    sr_val_data_t data;
     ...;
 } sr_val_t;
 typedef enum sr_edit_flag_e {
@@ -159,6 +167,12 @@ typedef enum sr_get_oper_flag_e {
 } sr_get_oper_flag_t;
 typedef uint32_t sr_get_oper_options_t;
 
+struct sr_data_s {
+    const sr_conn_ctx_t *conn;
+    struct lyd_node *tree;
+};
+typedef struct sr_data_s sr_data_t;
+
 void sr_free_val(sr_val_t *);
 void sr_free_values(sr_val_t *, size_t);
 
@@ -166,26 +180,25 @@ int sr_get_item(sr_session_ctx_t *, const char *, uint32_t, sr_val_t **);
 int sr_get_items(sr_session_ctx_t *, const char *, uint32_t,
 	const sr_get_oper_options_t, sr_val_t **, size_t *);
 int sr_get_data(sr_session_ctx_t *, const char *, uint32_t, uint32_t,
-	const sr_get_oper_options_t, struct lyd_node **);
-int sr_rpc_send_tree(
-	sr_session_ctx_t *, struct lyd_node *, uint32_t, struct lyd_node **);
+	const sr_get_oper_options_t, sr_data_t **);
+void sr_release_data(sr_data_t *);
+int sr_rpc_send_tree(sr_session_ctx_t *, struct lyd_node *, uint32_t, sr_data_t **);
 
 int sr_set_item_str(sr_session_ctx_t *, const char *, const char *, const char *, const sr_edit_options_t);
 int sr_delete_item(sr_session_ctx_t *, const char *, const sr_edit_options_t);
 int sr_edit_batch(sr_session_ctx_t *, const struct lyd_node *, const char *);
-int sr_replace_config(sr_session_ctx_t *, const char *, struct lyd_node *, uint32_t, int);
+int sr_replace_config(sr_session_ctx_t *, const char *, struct lyd_node *, uint32_t);
 int sr_validate(sr_session_ctx_t *, const char *, uint32_t);
-int sr_apply_changes(sr_session_ctx_t *, uint32_t, int);
+int sr_apply_changes(sr_session_ctx_t *, uint32_t);
 int sr_discard_changes(sr_session_ctx_t *);
 
 typedef enum sr_subscr_flag_e {
-	SR_SUBSCR_CTX_REUSE,
 	SR_SUBSCR_NO_THREAD,
 	SR_SUBSCR_PASSIVE,
 	SR_SUBSCR_DONE_ONLY,
 	SR_SUBSCR_ENABLED,
 	SR_SUBSCR_UPDATE,
-	SR_SUBSCR_UNLOCKED,
+	SR_SUBSCR_OPER_MERGE,
 	...
 } sr_subscr_flag_t;
 
@@ -193,8 +206,7 @@ typedef struct sr_subscription_ctx_s sr_subscription_ctx_t;
 typedef uint32_t sr_subscr_options_t;
 
 int sr_get_event_pipe(sr_subscription_ctx_t *, int *);
-typedef int... time_t;
-int sr_process_events(sr_subscription_ctx_t *, sr_session_ctx_t *, time_t *);
+int sr_subscription_process_events(sr_subscription_ctx_t *, sr_session_ctx_t *, struct timespec *);
 int sr_unsubscribe(sr_subscription_ctx_t *);
 
 typedef enum sr_event_e {
@@ -221,50 +233,50 @@ int sr_get_change_next(
 int sr_get_change_tree_next(
 	sr_session_ctx_t *, sr_change_iter_t *, sr_change_oper_t *,
 	const struct lyd_node **, const char **prev_val,
-	const char **prev_list, bool *prev_dflt);
+	const char **prev_list, int *prev_dflt);
 void sr_free_change_iter(sr_change_iter_t *);
 
 
 extern "Python" int srpy_module_change_cb(
-	sr_session_ctx_t *, const char *module, const char *xpath,
+	sr_session_ctx_t *, uint32_t sub_id, const char *module, const char *xpath,
 	sr_event_t, uint32_t req_id, void *priv);
 int sr_module_change_subscribe(
 	sr_session_ctx_t *, const char *module, const char *xpath,
-	int (*)(sr_session_ctx_t *, const char *, const char *, sr_event_t, uint32_t, void *),
+	int (*)(sr_session_ctx_t *, uint32_t, const char *, const char *, sr_event_t, uint32_t, void *),
 	void *priv, uint32_t priority, sr_subscr_options_t, sr_subscription_ctx_t **);
 
 extern "Python" int srpy_rpc_tree_cb(
-	sr_session_ctx_t *, const char *, const struct lyd_node *input,
+	sr_session_ctx_t *, uint32_t sub_id, const char *, const struct lyd_node *input,
 	sr_event_t, uint32_t req_id, struct lyd_node *output, void *priv);
 int sr_rpc_subscribe_tree(
 	sr_session_ctx_t *, const char *xpath,
-	int (*)(sr_session_ctx_t *, const char *, const struct lyd_node *, sr_event_t, uint32_t, struct lyd_node *, void *),
+	int (*)(sr_session_ctx_t *, uint32_t, const char *, const struct lyd_node *, sr_event_t, uint32_t, struct lyd_node *, void *),
 	void *priv, uint32_t priority, sr_subscr_options_t, sr_subscription_ctx_t **);
 
 extern "Python" int srpy_oper_data_cb(
-	sr_session_ctx_t *, const char *module, const char *xpath,
+	sr_session_ctx_t *, uint32_t sub_id, const char *module, const char *xpath,
 	const char *req_xpath, uint32_t req_id, struct lyd_node **, void *priv);
-int sr_oper_get_items_subscribe(
+int sr_oper_get_subscribe(
 	sr_session_ctx_t *, const char *module, const char *xpath,
-	int (*)(sr_session_ctx_t *, const char *, const char *, const char *, uint32_t, struct lyd_node **, void *),
+	int (*)(sr_session_ctx_t *, uint32_t, const char *, const char *, const char *, uint32_t, struct lyd_node **, void *),
 	void *priv, sr_subscr_options_t, sr_subscription_ctx_t **);
 
 typedef enum sr_ev_notif_type_e {
 	SR_EV_NOTIF_REALTIME,
 	SR_EV_NOTIF_REPLAY,
 	SR_EV_NOTIF_REPLAY_COMPLETE,
-	SR_EV_NOTIF_STOP,
+	SR_EV_NOTIF_TERMINATED,
 	SR_EV_NOTIF_SUSPENDED,
 	SR_EV_NOTIF_RESUMED,
 	...
 } sr_ev_notif_type_t;
 
 extern "Python" void srpy_event_notif_tree_cb(
-	sr_session_ctx_t *, const sr_ev_notif_type_t notif_type, const struct lyd_node *notif,
-	time_t timestamp, void *priv);
+	sr_session_ctx_t *, uint32_t sub_id, const sr_ev_notif_type_t notif_type, const struct lyd_node *notif,
+	struct timespec* timestamp, void *priv);
 
-int sr_event_notif_subscribe_tree(sr_session_ctx_t *, const char *module_name, const char *xpath, time_t start_time, time_t stop_time,
-	void (*)(sr_session_ctx_t *, const sr_ev_notif_type_t, const struct lyd_node*, time_t, void*),
+int sr_notif_subscribe_tree(sr_session_ctx_t *, const char *module_name, const char *xpath, struct timespec *start_time, struct timespec *stop_time,
+	void (*)(sr_session_ctx_t *, uint32_t, const sr_ev_notif_type_t, const struct lyd_node*, struct timespec*, void*),
 	void *priv, sr_subscr_options_t, sr_subscription_ctx_t **);
 
-int sr_event_notif_send_tree(sr_session_ctx_t *, struct lyd_node *notif);
+int sr_notif_send_tree(sr_session_ctx_t *, struct lyd_node *notif, uint32_t timeout_ms, int wait);
